@@ -66,7 +66,7 @@
 #define IO_CAPABILITY_DISPLAYONLY	0x00
 #define IO_CAPABILITY_DISPLAYYESNO	0x01
 #define IO_CAPABILITY_KEYBOARDONLY	0x02
-#define IO_CAPABILITY_NOINPUTOUTPUT	0x03
+#define IO_CAPABILITY_NOINPUTNOOUTPUT	0x03
 #define IO_CAPABILITY_INVALID		0xFF
 
 #define check_address(address) bachk(address)
@@ -74,6 +74,8 @@
 static DBusConnection *connection = NULL;
 static GSList *adapter_drivers = NULL;
 
+const struct btd_adapter_ops *adapter_ops = NULL;
+
 struct session_req {
 	struct btd_adapter	*adapter;
 	DBusConnection		*conn;		/* Connection reference */
@@ -250,45 +252,6 @@ int pending_remote_name_cancel(struct btd_adapter *adapter)
 	return err;
 }
 
-static int set_limited_discoverable(int dd, const uint8_t *cls,
-							gboolean limited)
-{
-	uint32_t dev_class;
-	int num = (limited ? 2 : 1);
-	uint8_t lap[] = { 0x33, 0x8b, 0x9e, 0x00, 0x8b, 0x9e };
-	/*
-	 * 1: giac
-	 * 2: giac + liac
-	 */
-	if (hci_write_current_iac_lap(dd, num, lap, HCI_REQ_TIMEOUT) < 0) {
-		int err = -errno;
-		error("Can't write current IAC LAP: %s(%d)",
-						strerror(errno), errno);
-		return err;
-	}
-
-	if (limited) {
-		if (cls[1] & 0x20)
-			return 0; /* Already limited */
-
-		dev_class = (cls[2] << 16) | ((cls[1] | 0x20) << 8) | cls[0];
-	} else {
-		if (!(cls[1] & 0x20))
-			return 0; /* Already clear */
-
-		dev_class = (cls[2] << 16) | ((cls[1] & 0xdf) << 8) | cls[0];
-	}
-
-	if (hci_write_class_of_dev(dd, dev_class, HCI_REQ_TIMEOUT) < 0) {
-		int err = -errno;
-		error("Can't write class of device: %s (%d)",
-						strerror(errno), errno);
-		return err;
-	}
-
-	return 0;
-}
-
 static const char *mode2str(uint8_t mode)
 {
 	switch(mode) {
@@ -341,23 +304,10 @@ static void adapter_remove_discov_timeout(struct btd_adapter *adapter)
 static gboolean discov_timeout_handler(gpointer user_data)
 {
 	struct btd_adapter *adapter = user_data;
-	int dd;
-	uint8_t scan_enable;
-	uint16_t dev_id = adapter->dev_id;
 
 	adapter->discov_timeout_id = 0;
 
-	dd = hci_open_dev(dev_id);
-	if (dd < 0) {
-		error("HCI device open failed: hci%d", dev_id);
-		return FALSE;
-	}
-
-	scan_enable = adapter->scan_mode & ~SCAN_INQUIRY;
-
-	hci_send_cmd(dd, OGF_HOST_CTL, OCF_WRITE_SCAN_ENABLE,
-					1, &scan_enable);
-	hci_close_dev(dd);
+	adapter_ops->set_connectable(adapter->dev_id);
 
 	return FALSE;
 }
@@ -378,98 +328,58 @@ static void adapter_set_discov_timeout(struct btd_adapter *adapter,
 							adapter);
 }
 
-static uint8_t mode2scan(uint8_t mode)
-{
-	switch (mode) {
-	case MODE_OFF:
-		return SCAN_DISABLED;
-	case MODE_CONNECTABLE:
-		return SCAN_PAGE;
-	case MODE_DISCOVERABLE:
-	case MODE_LIMITED:
-		return (SCAN_PAGE | SCAN_INQUIRY);
-	default:
-		error("Invalid mode given to mode2scan: %u", mode);
-		return SCAN_PAGE;
-	}
-}
-
 static int set_mode(struct btd_adapter *adapter, uint8_t new_mode)
 {
-	uint8_t scan_enable;
-	uint8_t current_scan = adapter->scan_mode;
-	int err, dd;
+	int err;
 	const char *modestr;
 
-	scan_enable = mode2scan(new_mode);
-
-	dd = hci_open_dev(adapter->dev_id);
-	if (dd < 0)
-		return -EIO;
-
-	if (!adapter->up && scan_enable != SCAN_DISABLED) {
-		/* Start HCI device */
-		if (ioctl(dd, HCIDEVUP, adapter->dev_id) == 0)
-			goto done; /* on success */
-
-		if (errno != EALREADY) {
-			err = -errno;
-			error("Can't init device hci%d: %s (%d)\n",
-				adapter->dev_id, strerror(errno), errno);
-
-			hci_close_dev(dd);
+	if (!adapter->up && new_mode != MODE_OFF) {
+		err = adapter_ops->set_powered(adapter->dev_id, TRUE);
+		if (err < 0)
 			return err;
-		}
 	}
 
-	if (adapter->up && scan_enable == SCAN_DISABLED) {
-		struct hci_request rq = {
-			.ogf = OGF_HOST_CTL,
-			.ocf = OCF_WRITE_SCAN_ENABLE,
-			.cparam = &scan_enable,
-			.clen = sizeof(scan_enable),
-		};
-
-		hci_send_req(dd, &rq, HCI_REQ_TIMEOUT);
-
-		if (ioctl(dd, HCIDEVDOWN, adapter->dev_id) < 0) {
-			err = -errno;
-			hci_close_dev(dd);
+	if (adapter->up && new_mode == MODE_OFF) {
+		err = adapter_ops->set_powered(adapter->dev_id, FALSE);
+		if (err < 0)
 			return err;
-		}
 
 		adapter->off_requested = TRUE;
 
 		goto done;
 	}
 
-	if (current_scan != scan_enable) {
-		err = hci_send_cmd(dd, OGF_HOST_CTL, OCF_WRITE_SCAN_ENABLE,
-					1, &scan_enable);
-		if (err < 0) {
-			hci_close_dev(dd);
-			return err;
-		}
-	} else if ((scan_enable & SCAN_INQUIRY) &&
-						(new_mode != adapter->mode)) {
+	if (new_mode == adapter->mode)
+		return 0;
+
+	if (new_mode == MODE_CONNECTABLE)
+		err = adapter_ops->set_connectable(adapter->dev_id);
+	else
+		err = adapter_ops->set_discoverable(adapter->dev_id);
+
+	if (err < 0)
+		return err;
+
+	if (new_mode > MODE_CONNECTABLE) {
 		adapter_remove_discov_timeout(adapter);
+
 		if (adapter->discov_timeout)
 			adapter_set_discov_timeout(adapter,
 						adapter->discov_timeout);
+
 		if (new_mode == MODE_LIMITED)
-			set_limited_discoverable(dd, adapter->dev.class,
-									TRUE);
+			adapter_ops->set_limited_discoverable(adapter->dev_id,
+						adapter->dev.class, TRUE);
 		else if (adapter->mode == MODE_LIMITED)
-			set_limited_discoverable(dd, adapter->dev.class,
-									FALSE);
+			adapter_ops->set_limited_discoverable(adapter->dev_id,
+						adapter->dev.class,FALSE);
 	}
+
 done:
 	modestr = mode2str(new_mode);
 
 	write_device_mode(&adapter->bdaddr, modestr);
 
-	hci_close_dev(dd);
-
 	adapter->mode = new_mode;
 
 	return 0;
@@ -640,6 +550,9 @@ static void session_remove(struct session_req *req)
 		g_slist_free(adapter->found_devices);
 		adapter->found_devices = NULL;
 
+		g_slist_free(adapter->oor_devices);
+		adapter->oor_devices = NULL;
+
 		if (adapter->state & STD_INQUIRY)
 			cancel_discovery(adapter);
 		else if (adapter->scheduler_id)
@@ -1077,7 +990,8 @@ static int start_inquiry(struct btd_adapter *adapter)
 
 	hci_close_dev(dd);
 
-	adapter->state |= RESOLVE_NAME;
+	if (main_opts.name_resolv)
+		adapter->state |= RESOLVE_NAME;
 
 	return 0;
 }
@@ -1127,7 +1041,8 @@ static int start_periodic_inquiry(struct btd_adapter *adapter)
 
 	hci_close_dev(dd);
 
-	adapter->state |= RESOLVE_NAME;
+	if (main_opts.name_resolv)
+		adapter->state |= RESOLVE_NAME;
 
 	return 0;
 }
@@ -1152,7 +1067,7 @@ static DBusMessage *adapter_start_discovery(DBusConnection *conn,
 	if (adapter->disc_sessions)
 		goto done;
 
-	if (main_opts.inqmode)
+	if (main_opts.discov_interval)
 		err = start_inquiry(adapter);
 	else
 		err = start_periodic_inquiry(adapter);
@@ -1483,9 +1398,15 @@ static DBusMessage *cancel_device_creation(DBusConnection *conn,
 				ERROR_INTERFACE ".NotInProgress",
 				"Device creation not in progress");
 
-	if (!device_is_bonding(device, sender))
+	if (device_is_bonding(device, NULL) &&
+			!device_is_bonding(device, sender))
 		return not_authorized(msg);
 
+	if (device_is_connected(device)) {
+		device_request_disconnect(device, msg);
+		return NULL;
+	}
+
 	adapter_remove_device(conn, adapter, device);
 
 	return dbus_message_new_method_return(msg);
@@ -1516,8 +1437,6 @@ static DBusMessage *create_device(DBusConnection *conn,
 	if (!device)
 		return NULL;
 
-	device_set_temporary(device, FALSE);
-
 	device_browse(device, conn, msg, NULL, FALSE);
 
 	return NULL;
@@ -1533,8 +1452,8 @@ static uint8_t parse_io_capability(const char *capability)
 		return IO_CAPABILITY_DISPLAYYESNO;
 	if (g_str_equal(capability, "KeyboardOnly"))
 		return IO_CAPABILITY_KEYBOARDONLY;
-	if (g_str_equal(capability, "NoInputOutput"))
-		return IO_CAPABILITY_NOINPUTOUTPUT;
+	if (g_str_equal(capability, "NoInputNoOutput"))
+		return IO_CAPABILITY_NOINPUTNOOUTPUT;
 	return IO_CAPABILITY_INVALID;
 }
 
@@ -1567,6 +1486,10 @@ static DBusMessage *create_paired_device(DBusConnection *conn,
 		return invalid_args(msg);
 
 	device = adapter_get_device(conn, adapter, address);
+	if (!device)
+		return g_dbus_create_error(msg,
+				ERROR_INTERFACE ".Failed",
+				"Unable to create a new device object");
 
 	return device_create_bonding(device, conn, msg, agent_path, cap);
 }
@@ -1603,13 +1526,15 @@ static DBusMessage *remove_device(DBusConnection *conn,
 				ERROR_INTERFACE ".DoesNotExist",
 				"Device creation in progress");
 
-	if (device_is_connected(device)) {
-		device_set_temporary(device, TRUE);
-		device_disconnect(device);
-	} else
-		adapter_remove_device(conn, adapter, device);
+	device_set_temporary(device, TRUE);
 
-	return dbus_message_new_method_return(msg);
+	if (!device_is_connected(device)) {
+		adapter_remove_device(conn, adapter, device);
+		return dbus_message_new_method_return(msg);
+	}
+
+	device_request_disconnect(device, msg);
+	return NULL;
 }
 
 static DBusMessage *find_device(DBusConnection *conn,
@@ -1730,8 +1655,10 @@ static GDBusMethodTable adapter_methods[] = {
 						G_DBUS_METHOD_FLAG_ASYNC},
 	{ "CreatePairedDevice",	"sos",	"o",	create_paired_device,
 						G_DBUS_METHOD_FLAG_ASYNC},
-	{ "CancelDeviceCreation","s",	"",	cancel_device_creation	},
-	{ "RemoveDevice",	"o",	"",	remove_device		},
+	{ "CancelDeviceCreation","s",	"",	cancel_device_creation,
+						G_DBUS_METHOD_FLAG_ASYNC},
+	{ "RemoveDevice",	"o",	"",	remove_device,
+						G_DBUS_METHOD_FLAG_ASYNC},
 	{ "FindDevice",		"s",	"o",	find_device		},
 	{ "RegisterAgent",	"os",	"",	register_agent		},
 	{ "UnregisterAgent",	"o",	"",	unregister_agent	},
@@ -1808,7 +1735,7 @@ static int adapter_setup(struct btd_adapter *adapter, int dd)
 	int err;
 	char name[249];
 
-	if (dev->hci_rev > 1) {
+	if (dev->lmp_ver > 1) {
 		if (dev->features[5] & LMP_SNIFF_SUBR)
 			events[5] |= 0x20;
 
@@ -1955,11 +1882,15 @@ static void load_drivers(struct btd_adapter *adapter)
 	}
 }
 
-static void load_connections(struct btd_adapter *adapter, int dd)
+static void load_connections(struct btd_adapter *adapter)
 {
 	struct hci_conn_list_req *cl = NULL;
 	struct hci_conn_info *ci;
-	int i;
+	int i, dd;
+
+	dd = hci_open_dev(adapter->dev_id);
+	if (dd < 0)
+		return;
 
 	cl = g_malloc0(10 * sizeof(*ci) + sizeof(*cl));
 
@@ -1969,6 +1900,7 @@ static void load_connections(struct btd_adapter *adapter, int dd)
 
 	if (ioctl(dd, HCIGETCONNLIST, cl) != 0) {
 		g_free(cl);
+		hci_close_dev(dd);
 		return;
 	}
 
@@ -1983,6 +1915,7 @@ static void load_connections(struct btd_adapter *adapter, int dd)
 	}
 
 	g_free(cl);
+	hci_close_dev(dd);
 }
 
 static int get_discoverable_timeout(const char *src)
@@ -2005,11 +1938,12 @@ static int get_pairable_timeout(const char *src)
 	return main_opts.pairto;
 }
 
-static int adapter_up(struct btd_adapter *adapter, int dd)
+static int adapter_up(struct btd_adapter *adapter)
 {
 	char mode[14], srcaddr[18];
 	uint8_t scan_mode;
 	gboolean powered, dev_down = FALSE;
+	int err;
 
 	ba2str(&adapter->bdaddr, srcaddr);
 
@@ -2052,7 +1986,7 @@ static int adapter_up(struct btd_adapter *adapter, int dd)
 		else
 			write_device_mode(&adapter->bdaddr, mode);
 
-		return adapter_up(adapter, dd);
+		return adapter_up(adapter);
 	} else if (!g_str_equal(mode, "connectable") &&
 			adapter->discov_timeout == 0) {
 		/* Set discoverable only if timeout is 0 */
@@ -2062,9 +1996,13 @@ static int adapter_up(struct btd_adapter *adapter, int dd)
 	}
 
 proceed:
-	if (dev_down == FALSE)
-		hci_send_cmd(dd, OGF_HOST_CTL, OCF_WRITE_SCAN_ENABLE,
-				1, &scan_mode);
+	if (scan_mode == SCAN_PAGE)
+		err = adapter_ops->set_connectable(adapter->dev_id);
+	else
+		err = adapter_ops->set_discoverable(adapter->dev_id);
+
+	if (err < 0)
+		return err;
 
 	if (adapter->initialized == FALSE) {
 		load_drivers(adapter);
@@ -2072,17 +2010,19 @@ proceed:
 
 		/* retrieve the active connections: address the scenario where
 		 * the are active connections before the daemon've started */
-		load_connections(adapter, dd);
+		load_connections(adapter);
 
 		adapter->initialized = TRUE;
 
-		adapter_update(adapter, 0, FALSE);
-
 		manager_add_adapter(adapter->path);
+
 	}
 
+	if (adapter->svc_cache)
+		adapter_update(adapter, 0);
+
 	if (dev_down) {
-		ioctl(dd, HCIDEVDOWN, adapter->dev_id);
+		adapter_ops->stop(adapter->dev_id);
 		adapter->off_requested = TRUE;
 		return 1;
 	} else
@@ -2203,10 +2143,10 @@ setup:
 	adapter->state &= ~STD_INQUIRY;
 
 	adapter_setup(adapter, dd);
-	err = adapter_up(adapter, dd);
-
 	hci_close_dev(dd);
 
+	err = adapter_up(adapter);
+
 	info("Adapter %s has been enabled", adapter->path);
 
 	return err;
@@ -2319,8 +2259,7 @@ int adapter_stop(struct btd_adapter *adapter)
 	return 0;
 }
 
-int adapter_update(struct btd_adapter *adapter, uint8_t new_svc,
-							gboolean starting)
+int adapter_update(struct btd_adapter *adapter, uint8_t new_svc)
 {
 	struct hci_dev *dev = &adapter->dev;
 	int dd;
@@ -2329,7 +2268,7 @@ int adapter_update(struct btd_adapter *adapter, uint8_t new_svc,
 	if (dev->ignore)
 		return 0;
 
-	if (starting || !adapter->initialized) {
+	if (!adapter->up) {
 		adapter->svc_cache = new_svc;
 		return 0;
 	}
@@ -2434,7 +2373,8 @@ struct btd_adapter *adapter_create(DBusConnection *conn, int id,
 	}
 
 	adapter->dev_id = id;
-	adapter->state |= RESOLVE_NAME;
+	if (main_opts.name_resolv)
+		adapter->state |= RESOLVE_NAME;
 	adapter->path = g_strdup(path);
 	adapter->already_up = devup;
 
@@ -2456,23 +2396,16 @@ void adapter_remove(struct btd_adapter *adapter)
 
 	debug("Removing adapter %s", path);
 
-	unload_drivers(adapter);
-
 	for (l = adapter->devices; l; l = l->next)
 		device_remove(l->data, connection, FALSE);
 	g_slist_free(adapter->devices);
 
+	unload_drivers(adapter);
+
 	/* Return adapter to down state if it was not up on init */
-	if (adapter->up && !adapter->already_up) {
-		int dd = hci_open_dev(adapter->dev_id);
-		if (dd < 0)
-			goto done;
+	if (adapter->up && !adapter->already_up)
+		adapter_ops->stop(adapter->dev_id);
 
-		ioctl(dd, HCIDEVDOWN, adapter->dev_id);
-		hci_close_dev(dd);
-	}
-
-done:
 	g_dbus_unregister_interface(connection, path, ADAPTER_INTERFACE);
 	g_free(path);
 }
@@ -2505,8 +2438,9 @@ void adapter_set_state(struct btd_adapter *adapter, int state)
 
 	if (state & PERIODIC_INQUIRY || state & STD_INQUIRY)
 		discov_active = TRUE;
-	else if (adapter->disc_sessions && main_opts.inqmode)
-		adapter->scheduler_id = g_timeout_add_seconds(main_opts.inqmode,
+	else if (adapter->disc_sessions && main_opts.discov_interval)
+		adapter->scheduler_id = g_timeout_add_seconds(
+						main_opts.discov_interval,
 						(GSourceFunc) start_inquiry,
 						adapter);
 
@@ -2660,6 +2594,11 @@ void adapter_update_found_devices(struct btd_adapter *adapter, bdaddr_t *bdaddr,
 	if (dev) {
 		if (rssi == dev->rssi)
 			return;
+
+		/* Out of range list update */
+		adapter->oor_devices = g_slist_remove(adapter->oor_devices,
+							dev);
+
 		goto done;
 	}
 
@@ -2704,53 +2643,27 @@ int adapter_remove_found_device(struct btd_adapter *adapter, bdaddr_t *bdaddr)
 void adapter_update_oor_devices(struct btd_adapter *adapter)
 {
 	GSList *l;
-	struct remote_dev_info *dev;
-	bdaddr_t tmp;
 
 	for (l = adapter->oor_devices; l; l = l->next) {
-		char *address = l->data;
-		struct remote_dev_info match;
+		char address[18];
+		const char *paddr = address;
+		struct remote_dev_info *dev = l->data;
 
-		memset(&match, 0, sizeof(struct remote_dev_info));
-		str2ba(address, &match.bdaddr);
+		ba2str(&dev->bdaddr, address);
 
 		g_dbus_emit_signal(connection, adapter->path,
 				ADAPTER_INTERFACE, "DeviceDisappeared",
-				DBUS_TYPE_STRING, &address,
+				DBUS_TYPE_STRING, &paddr,
 				DBUS_TYPE_INVALID);
 
-		dev = adapter_search_found_devices(adapter, &match);
-		if (!dev)
-			continue;
-
 		adapter->found_devices = g_slist_remove(adapter->found_devices, dev);
 		dev_info_free(dev);
 	}
 
-	g_slist_foreach(adapter->oor_devices, (GFunc) free, NULL);
 	g_slist_free(adapter->oor_devices);
 	adapter->oor_devices = NULL;
 
-	for (l = adapter->found_devices; l; l = l->next) {
-		dev = l->data;
-		baswap(&tmp, &dev->bdaddr);
-		adapter->oor_devices = g_slist_append(adapter->oor_devices,
-							batostr(&tmp));
-	}
-}
-
-void adapter_remove_oor_device(struct btd_adapter *adapter, char *peer_addr)
-{
-	GSList *l;
-
-	l = g_slist_find_custom(adapter->oor_devices, peer_addr,
-				(GCompareFunc) strcmp);
-	if (l) {
-		char *dev = l->data;
-		adapter->oor_devices = g_slist_remove(adapter->oor_devices,
-								dev);
-		g_free(dev);
-	}
+	adapter->oor_devices = g_slist_copy(adapter->found_devices);
 }
 
 void adapter_mode_changed(struct btd_adapter *adapter, uint8_t scan_mode)
@@ -2758,7 +2671,6 @@ void adapter_mode_changed(struct btd_adapter *adapter, uint8_t scan_mode)
 	const gchar *path = adapter_get_path(adapter);
 	gboolean discoverable, pairable;
 	uint8_t real_class[3];
-	int dd;
 
 	if (adapter->scan_mode == scan_mode)
 		return;
@@ -2803,23 +2715,17 @@ void adapter_mode_changed(struct btd_adapter *adapter, uint8_t scan_mode)
 					ADAPTER_INTERFACE, "Pairable",
 					DBUS_TYPE_BOOLEAN, &pairable);
 
-	dd = hci_open_dev(adapter->dev_id);
-	if (dd < 0) {
-		error("HCI device open failed: hci%d", adapter->dev_id);
-		goto done;
-	}
-
 	memcpy(real_class, adapter->dev.class, 3);
 	if (adapter->svc_cache)
 		real_class[2] = adapter->svc_cache;
 
 	if (discoverable && adapter->pairable)
-		set_limited_discoverable(dd, real_class, TRUE);
+		adapter_ops->set_limited_discoverable(adapter->dev_id,
+							real_class, TRUE);
 	else if (!discoverable)
-		set_limited_discoverable(dd, real_class, FALSE);
+		adapter_ops->set_limited_discoverable(adapter->dev_id,
+							real_class, FALSE);
 
-	hci_close_dev(dd);
-done:
 	emit_property_changed(connection, path,
 				ADAPTER_INTERFACE, "Discoverable",
 				DBUS_TYPE_BOOLEAN, &discoverable);
@@ -3084,3 +2990,30 @@ gboolean adapter_powering_down(struct btd_adapter *adapter)
 {
 	return adapter->off_requested;
 }
+
+int btd_register_adapter_ops(struct btd_adapter_ops *btd_adapter_ops)
+{
+	/* Already registered */
+	if (adapter_ops)
+		return -EALREADY;
+
+	if (btd_adapter_ops->setup == NULL)
+		return -EINVAL;
+
+	adapter_ops = btd_adapter_ops;
+
+	return 0;
+}
+
+void btd_adapter_cleanup_ops()
+{
+	adapter_ops->cleanup();
+}
+
+int adapter_ops_setup()
+{
+	if (!adapter_ops)
+		return -EINVAL;
+
+	return adapter_ops->setup();
+}
